import math

from pygments import highlight  # pyright: ignore[reportUnknownVariableType]
from pygments.formatters import TerminalFormatter
from termcolor import colored

from .code_change import CodeChange, CodeChangeAction

change_delimiter = 60 * "="


def _remove_extra_empty_lines(lines: list[str]) -> list[str]:
    if not lines:
        return []

    # Find the first non-empty line
    start = 0
    while start < len(lines) and not lines[start].strip():
        start += 1

    # Find the last non-empty line
    end = len(lines) - 1
    while end > start and not lines[end].strip():
        end -= 1

    # If all lines are empty, keep only one empty line
    if start == len(lines):
        return [" "]

    # Return the list with only a maximum of one empty line on either side
    return lines[max(start - 1, 0) : end + 2]


def _prefixed_lines(code_change: CodeChange, lines: list[str], prefix: str):
    return "\n".join(
        [
            prefix
            + " " * (code_change.line_number_buffer - len(prefix))
            + line.strip("\n")
            for line in lines
        ]
    )


def print_change(code_change: CodeChange):
    to_print = [
        get_file_name(code_change),
        change_delimiter if code_change.action != CodeChangeAction.RenameFile else "",
        get_previous_lines(code_change),
        get_removed_block(code_change),
        get_added_block(code_change),
        get_later_lines(code_change),
        change_delimiter if code_change.action != CodeChangeAction.RenameFile else "",
    ]
    for s in to_print:
        if s:
            print(s)


def get_file_name(code_change: CodeChange):
    file_name = code_change.file
    match code_change.action:
        case CodeChangeAction.CreateFile:
            return colored(f"\n{file_name}*", color="light_green")
        case CodeChangeAction.DeleteFile:
            return colored(f"\n{file_name}", color="light_red")
        case CodeChangeAction.RenameFile:
            return colored(
                f"\nRename: {file_name} -> {code_change.name}", color="yellow"
            )
        case _:
            return colored(f"\n{file_name}", color="light_blue")


def get_removed_block(
    code_change: CodeChange, prefix: str = "-", color: str | None = "red"
):
    if code_change.action.has_removals():
        if code_change.action == CodeChangeAction.DeleteFile:
            changed_lines = code_change.file_lines
        else:
            changed_lines = code_change.file_lines[
                code_change.first_changed_line - 1 : code_change.last_changed_line
            ]

        removed = _prefixed_lines(code_change, changed_lines, prefix)
        if removed:
            return colored(removed, color=color)
    return ""


def get_added_block(
    code_change: CodeChange, prefix: str = "+", color: str | None = "green"
):
    if code_change.action.has_additions():
        added = _prefixed_lines(code_change, code_change.code_lines, prefix)
        if added:
            return colored(added, color=color)
    return ""


def get_previous_lines(code_change: CodeChange, num: int = 2):
    if not code_change.action.has_surrounding_lines():
        return ""
    lines = _remove_extra_empty_lines(
        [
            code_change.file_lines[i]
            for i in range(
                max(0, math.ceil(code_change.first_changed_line) - (num + 1)),
                min(
                    math.ceil(code_change.first_changed_line) - 1,
                    len(code_change.file_lines),
                ),
            )
        ]
    )
    numbered = [
        (str(math.ceil(code_change.first_changed_line) - len(lines) + i) + ":").ljust(
            code_change.line_number_buffer
        )
        + line
        for i, line in enumerate(lines)
    ]

    prev = "\n".join(numbered)
    if prev:
        # pygments doesn't have type hints on TerminalFormatter
        h_prev: str = highlight(prev, code_change.lexer, TerminalFormatter(bg="dark"))  # type: ignore
        return h_prev
    return ""


def get_later_lines(code_change: CodeChange, num: int = 2):
    if not code_change.action.has_surrounding_lines():
        return ""
    lines = _remove_extra_empty_lines(
        [
            code_change.file_lines[i]
            for i in range(
                max(0, int(code_change.last_changed_line)),
                min(
                    int(code_change.last_changed_line) + num,
                    len(code_change.file_lines),
                ),
            )
        ]
    )
    numbered = [
        (str(int(code_change.last_changed_line) + 1 + i) + ":").ljust(
            code_change.line_number_buffer
        )
        + line
        for i, line in enumerate(lines)
    ]

    later = "\n".join(numbered)
    if later:
        # pygments doesn't have type hints on TerminalFormatter
        h_later: str = highlight(later, code_change.lexer, TerminalFormatter(bg="dark"))  # type: ignore
        return h_later
    return ""



import os
from textwrap import dedent

from mentat.app import run


def test_insert(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Create a temporary file
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
                        # This is a temporary file
                        # with 2 lines"""))

    mock_collect_user_input.side_effect = [
        "Insert a comment between both lines",
        "y",
        KeyboardInterrupt,
    ]

    mock_call_llm_api.set_generator_values([dedent("""\
        I will insert a comment between both lines.

        Steps: 1. Insert a comment after the first line

        @@start
        {{
            "file": "{file_name}",
            "action": "insert",
            "insert-after-line": 1,
            "insert-before-line": 2
        }}
        @@code
        # I inserted this comment
        @@end""".format(file_name=temp_file_name))])

    # Run the system with the temporary file path
    run([temp_file_name])

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
                        # This is a temporary file
                        # I inserted this comment
                        # with 2 lines""")
    assert content == expected_content


def test_replace(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Create a temporary file
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
                        # This is a temporary file
                        # with 2 lines"""))

    mock_collect_user_input.side_effect = [
        "Replace both lines with one comment",
        "y",
        KeyboardInterrupt,
    ]

    mock_call_llm_api.set_generator_values([dedent("""\
        I will replace both lines with one comment

        Steps: 1. Replace both lines with one comment

        @@start
        {{
            "file": "{file_name}",
            "action": "replace",
            "start-line": 1,
            "end-line": 2
        }}
        @@code
        # I inserted this comment
        @@end""".format(file_name=temp_file_name))])

    # Run the system with the temporary file path
    run([temp_file_name])

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
                        # I inserted this comment""")
    assert content == expected_content


def test_delete(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Create a temporary file
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
                        # This is a temporary
                        # file
                        # with 4 
                        # lines"""))

    mock_collect_user_input.side_effect = [
        "Delete the middle two lines",
        "y",
        KeyboardInterrupt,
    ]

    mock_call_llm_api.set_generator_values([dedent("""\
        I will delete the middle two lines

        Steps: 1. Delete the middle two lines

        @@start
        {{
            "file": "{file_name}",
            "action": "delete",
            "start-line": 2,
            "end-line": 3
        }}
        @@end""".format(file_name=temp_file_name))])

    # Run the system with the temporary file path
    run([temp_file_name])

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
                        # This is a temporary
                        # lines""")
    assert content == expected_content


def test_create_file(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Create a temporary file
    temp_file_name = "new_dir/temp.py"
    mock_collect_user_input.side_effect = [
        "Create a new file called temp.py",
        "y",
        KeyboardInterrupt,
    ]

    mock_call_llm_api.set_generator_values([dedent("""\
        I will create a new file called temp.py

        Steps: 1. Create a new file called temp.py

        @@start
        {{
            "file": "{file_name}",
            "action": "create-file"
        }}
        @@code
        # I created this file
        @@end""".format(file_name=temp_file_name))])

    run(["."])

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
                        # I created this file""")
    assert content == expected_content


def test_delete_file(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Create a temporary file
    temp_file_name = "incredibly_temp.py"
    with open(temp_file_name, "w") as f:
        f.write("# I am not long for this world")

    mock_collect_user_input.side_effect = [
        "Delete the file",
        "y",
        "y",
        KeyboardInterrupt,
    ]

    mock_call_llm_api.set_generator_values([dedent("""\
        I will delete the file

        Steps: 1. Delete the file

        @@start
        {{
            "file": "{file_name}",
            "action": "delete-file"
        }}
        @@end""".format(file_name=temp_file_name))])

    # Run the system with the temporary file path
    run([temp_file_name])

    # Check if the temporary file is modified as expected
    assert not os.path.exists(temp_file_name)


def test_rename_file(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Make sure rename-file works
    temp_file_name = "temp.py"
    temp_2_file_name = "temp_2.py"
    with open(temp_file_name, "w") as f:
        f.write("# Move me!")

    mock_collect_user_input.side_effect = [
        "Rename the file temp_2.py",
        "y",
        KeyboardInterrupt,
    ]
    mock_call_llm_api.set_generator_values([dedent(f"""\
        I will rename the file

        Steps: 1. rename the file

        @@start
        {{
            "file": "{temp_file_name}",
            "action": "rename-file",
            "name": "{temp_2_file_name}"
        }}
        @@end""")])

    run([temp_file_name])
    with open(temp_2_file_name) as new_file:
        content = new_file.read()
        expected_content = "# Move me!"

    assert not os.path.exists(temp_file_name)
    assert content == expected_content


def test_change_then_rename_file(
    mock_call_llm_api, mock_collect_user_input, mock_setup_api_key
):
    # Make sure a change made before a rename works
    temp_file_name = "temp.py"
    temp_2_file_name = "temp_2.py"
    with open(temp_file_name, "w") as f:
        f.write("# Move me!")

    mock_collect_user_input.side_effect = [
        "Insert a comment then rename the file temp_2.py",
        "y",
        KeyboardInterrupt,
    ]
    mock_call_llm_api.set_generator_values([dedent(f"""\
        I will insert a comment then rename the file

        Steps:
        1. insert a comment
        2. rename the file

        @@start
        {{
            "file": "{temp_file_name}",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }}
        @@code
        # I inserted this comment!
        @@end
        @@start
        {{
            "file": "{temp_file_name}",
            "action": "rename-file",
            "name": "{temp_2_file_name}"
        }}
        @@end""")])

    run([temp_file_name])
    with open(temp_2_file_name) as new_file:
        content = new_file.read()
        expected_content = "# I inserted this comment!\n# Move me!"

    assert not os.path.exists(temp_file_name)
    assert content == expected_content


def test_rename_file_then_change(
    mock_call_llm_api, mock_collect_user_input, mock_setup_api_key
):
    # Make sure a change made after a rename works
    temp_file_name = "temp.py"
    temp_2_file_name = "temp_2.py"
    with open(temp_file_name, "w") as f:
        f.write("# Move me!")

    mock_collect_user_input.side_effect = [
        "Rename the file temp_2.py then insert a comment",
        "y",
        KeyboardInterrupt,
    ]
    mock_call_llm_api.set_generator_values([dedent(f"""\
        I will rename the file then insert a comment

        Steps:
        1. rename the file
        2. insert a comment

        @@start
        {{
            "file": "{temp_file_name}",
            "action": "rename-file",
            "name": "{temp_2_file_name}"
        }}
        @@end
        @@start
        {{
            "file": "{temp_2_file_name}",
            "action": "insert",
            "insert-after-line": 0,
            "insert-before-line": 1
        }}
        @@code
        # I inserted this comment!
        @@end""")])

    run([temp_file_name])
    with open(temp_2_file_name) as new_file:
        content = new_file.read()
        expected_content = "# I inserted this comment!\n# Move me!"

    assert not os.path.exists(temp_file_name)
    assert content == expected_content


def test_multiple_blocks(
    mock_call_llm_api, mock_collect_user_input, mock_setup_api_key
):
    # Create a temporary file
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
                        # This is a temporary file
                        # with 4 lines
                        # just for
                        # good measure"""))

    mock_collect_user_input.side_effect = [
        (
            "Insert a comment between the first 2 lines and then replace the last line"
            " with 'better measure'"
        ),
        "y",
        KeyboardInterrupt,
    ]

    mock_call_llm_api.set_generator_values([dedent("""\
        I will insert a comment between the first two lines
        and then replace the last line with 'better measure'

        Steps: 1. Insert a comment
               2. Replace last line

        @@start
        {{
            "file": "{file_name}",
            "action": "insert",
            "insert-after-line": 1,
            "insert-before-line": 2
        }}
        @@code
        # I inserted this comment
        @@end
        @@start
        {{
            "file": "{file_name}",
            "action": "replace",
            "start-line": 4,
            "end-line": 4
        }}
        @@code
        # better measure
        @@end""".format(file_name=temp_file_name))])

    # Run the system with the temporary file path
    run([temp_file_name])

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
                        # This is a temporary file
                        # I inserted this comment
                        # with 4 lines
                        # just for
                        # better measure""")
    assert content == expected_content


def test_json_strings(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Make sure we don't throw error if GPT gives us numbers in a string format
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write(dedent("""\
            # This is a temporary file"""))

    mock_collect_user_input.side_effect = [
        "Insert a comment at the start and replace the current line",
        "y",
        KeyboardInterrupt,
    ]

    mock_call_llm_api.set_generator_values([dedent("""\
        I will insert a comment at the start.

        @@start
        {{
            "file": "{file_name}",
            "action": "insert",
            "insert-after-line": "0",
            "insert-before-line": "1"
        }}
        @@code
        # I inserted this comment
        @@end
        @@start
        {{
            "file": "{file_name}",
            "action": "replace",
            "start-line": "1",
            "end-line": "1"
        }}
        @@code
        # I replaced this comment
        @@end""".format(file_name=temp_file_name))])

    # Run the system with the temporary file path
    run([temp_file_name])

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = dedent("""\
            # I inserted this comment
            # I replaced this comment""")
    assert content == expected_content



import glob
import logging
import os
from pathlib import Path
from typing import Any, Dict

from termcolor import cprint

from .code_file import CodeFile
from .config_manager import ConfigManager
from .errors import UserError
from .git_handler import get_non_gitignored_files, get_paths_with_git_diffs


def _is_file_text_encoded(file_path: Path):
    try:
        # The ultimate filetype test
        with open(file_path) as f:
            f.read()
        return True
    except UnicodeDecodeError:
        return False


def _abs_files_from_list(paths: list[Path], check_for_text: bool = True):
    files_direct = set[CodeFile]()
    file_paths_from_dirs = set[Path]()
    for path in paths:
        file = CodeFile(path)
        path = Path(file.path)
        if path.is_file():
            if check_for_text and not _is_file_text_encoded(path):
                logging.info(f"File path {path} is not text encoded.")
                raise UserError(f"File path {path} is not text encoded.")
            files_direct.add(file)
        elif path.is_dir():
            nonignored_files = set(
                map(
                    lambda f: Path(os.path.realpath(path / f)),
                    get_non_gitignored_files(path),
                )
            )

            file_paths_from_dirs.update(
                filter(
                    lambda f: (not check_for_text) or _is_file_text_encoded(f),
                    nonignored_files,
                )
            )

    files_from_dirs = [CodeFile(path) for path in file_paths_from_dirs]
    return files_direct, files_from_dirs


def _get_files(
    config: ConfigManager, paths: list[Path], exclude_paths: list[Path]
) -> Dict[Path, CodeFile]:
    excluded_files_direct, excluded_files_from_dirs = _abs_files_from_list(
        exclude_paths, check_for_text=False
    )
    excluded_files, excluded_files_from_dir = set(
        map(lambda f: f.path, excluded_files_direct)
    ), set(map(lambda f: f.path, excluded_files_from_dirs))

    glob_excluded_files = set(
        os.path.join(config.git_root, file)
        for glob_path in config.file_exclude_glob_list()
        # If the user puts a / at the beginning, it will try to look in root directory
        for file in glob.glob(
            pathname=glob_path,
            root_dir=config.git_root,
            recursive=True,
        )
    )
    files_direct, files_from_dirs = _abs_files_from_list(paths, check_for_text=True)

    # config glob excluded files only apply to files added from directories
    files_from_dirs = [
        file
        for file in files_from_dirs
        if str(file.path.resolve()) not in glob_excluded_files
    ]

    files_direct.update(files_from_dirs)

    files = dict[Path, CodeFile]()
    for file in files_direct:
        if file.path not in excluded_files | excluded_files_from_dir:
            files[Path(os.path.realpath(file.path))] = file

    return files


def _build_path_tree(files: list[CodeFile], git_root: Path):
    tree = dict[str, Any]()
    for file in files:
        path = os.path.relpath(file.path, git_root)
        parts = Path(path).parts
        current_level = tree
        for part in parts:
            if part not in current_level:
                current_level[part] = {}
            current_level = current_level[part]
    return tree


def _print_path_tree(
    tree: dict[str, Any], changed_files: set[Path], cur_path: Path, prefix: str = ""
):
    keys = list(tree.keys())
    for i, key in enumerate(sorted(keys)):
        if i < len(keys) - 1:
            new_prefix = prefix + "│   "
            print(f"{prefix}├── ", end="")
        else:
            new_prefix = prefix + "    "
            print(f"{prefix}└── ", end="")

        cur = cur_path / key
        star = "* " if cur in changed_files else ""
        if tree[key]:
            color = "blue"
        elif star:
            color = "green"
        else:
            color = None
        cprint(f"{star}{key}", color)
        if tree[key]:
            _print_path_tree(tree[key], changed_files, cur, new_prefix)


class CodeContext:
    def __init__(
        self,
        config: ConfigManager,
        paths: list[Path],
        exclude_paths: list[Path],
    ):
        self.config = config
        self.files = _get_files(self.config, paths, exclude_paths)

    def display_context(self):
        if self.files:
            cprint("Files included in context:", "green")
        else:
            cprint("No files included in context.\n", "red")
            cprint("Git project: ", "green", end="")
        cprint(self.config.git_root.name, "blue")
        _print_path_tree(
            _build_path_tree(list(self.files.values()), self.config.git_root),
            get_paths_with_git_diffs(self.config.git_root),
            self.config.git_root,
        )
        print()



from textwrap import dedent

system_prompt = """
    You are part of an automated coding system. As such, responses must adhere strictly
    to the required format, so they can be parsed programmaticaly. Your input will
    consist of a user request, the contents of code files, and sometimes the git diff of
    current code files. The request may be to add a new feature, update the code, fix a
    bug, add comments or docstrings, etc. The first part of your response should contain
    an brief summary of the changes you plan to make, then a list of the changes. Ensure
    you plan ahead, like planning to add imports for things you need to use in your
    changes, etc. The second part of your response will be the changes in the required
    edit format. Code edits consist of either inserts, deletes, replacements, creating
    new files, deleting existing files, or renaming existing files. They can be of multiple lines of code. Edit
    description blocks start with @@start and end with @@end. If the edit is a delete 
    or delete-file, then the block should only contain a JSON formatted section. In
    insert, replace, and create-file blocks, there must be a second section containing
    the new line or lines of code. The JSON section and code section are separated by
    a line containing just @@code. If the request requires clarification or the user is
    asking for something other than code changes, such as design ideas, don't return 
    any edit description blocks.


    Example 1:

    To demonstrate the response format, here's an example user request, followed by an example response:

    Code Files:

    core/script.py
    1:
    2:def say_hello(name):
    3:    print(f"Hello {name}!")
    4:
    5:
    6:def say_goodbye():
    7:    print("Goodbye!")
    8:
    9:
    10:def main(name):
    11:    say_hello(name)
    12:    say_goodbye()
    13:    print("Done!")
    14:

    core/hello_world.py
    1:
    2:def hello_world():
    3:    print("Hello, World!")
    4:

    Code Map:

    .gitignore

    core/hello_world.py
            function
                    hello_world ()

    core/script.py
            function
                    main (name)
                    say_goodbye ()
                    say_hello (name)

    User Request:
    After saying hello, if the user's name is "Bob", say "Nice to see you again!" on another line.
    Add a function to get the user's name and use it in main instead of taking name as an argument.
    The new function should be in a separate file called utils.py. Stop saying "Done!". Then, 
    rename hello_world.py to hello_again.py. Finally, delete the goodbye_world.py file.


    Example Response:

    I will make the modifications to script.py and create the new file, importing from it in script.py.

    Steps:
    1. Modify say_hello, adding the case for Bob.
    2. Create utils.py with a function to get the user's name.
    3. Import the new function in script.py.
    4. Modify main to use the new function instead of taking name as an argument.
    5. Remove the line printing "Done!".
    6. Rename the file hello_world.py to hello_again.py
    7. Delete file goodbye_world.py

    @@start
    {
        "file": "core/script.py",
        "action": "insert",
        "insert-after-line": 3,
        "insert-before-line": 4
    }
    @@code
        if name == "Bob":
            print("Nice to see you again!")
    @@end
    @@start
    {
        "file": "core/utils.py",
        "action": "create-file"
    }
    @@code
    def get_name():
        return input("Enter your name: ")
    @@end
    @@start
    {
        "file": "core/script.py",
        "action": "insert",
        "insert-after-line": 0,
        "insert-before-line": 1
    }
    @@code
    from core.utils import get_name
    @@end
    @@start
    {
        "file": "core/script.py",
        "action": "replace",
        "start-line": 10,
        "end-line": 10
    }
    @@code
    def main():
        name = get_name()
    @@end
    @@start
    {
        "file": "core/script.py",
        "action": "delete",
        "start-line": 13,
        "end-line": 13
    }
    @@end
    @@start
    {
        "file": "core/hello_world.py",
        "action": "rename-file",
        "name": "core/hello_again.py"
    }
    @@end
    @@start
    {
        "file": "core/goodbye_world.py",
        "action": "delete-file"
    }
    @@end


    Example 2:

    Here's an example response to a user request where the user is requesting changes to
    a function that is in "Code Files", but they want to use a function that needs to be 
    imported from a file in "Code Map":

    Code Files:

    core/hello_world.py
    1:
    2:def hello_world():
    3:    print("Hello, World!")
    4:

    Code Map:

    .gitignore

    core/hello_world.py
            function
                    hello_world ()

    core/script.py
            function
                    main (name)
                    say_goodbye ()
                    say_hello (name)

    User Request:
    Call say_goodbye after printing hello world

    Example Response:

    I will make the modifications to hello_world.py

    Steps:
    1. Import the say_goodbye function in hello_world.py
    2. Modify hello_world.py, adding a function call for say_goodbye

    @@start
    {
        "file": "core/hello_world.py",
        "action": "insert",
        "insert-after-line": 0,
        "insert-before-line": 1
    }
    @@code
    from core.script import say_goodbye
    @@end
    @@start
    {
        "file": "core/hello_world.py",
        "action": "insert",
        "insert-after-line": 4,
        "insert-before-line": 5
    }
    @@code
        say_goodbye()
    @@end
"""
system_prompt = dedent(system_prompt).strip()



import subprocess

from mentat.commands import Command, HelpCommand, InvalidCommand


def test_invalid_command():
    assert isinstance(Command.create_command("non-existent"), InvalidCommand)


def test_help_command():
    command = Command.create_command("help")
    command.apply()
    assert isinstance(command, HelpCommand)


def test_commit_command(temp_testbed):
    file_name = "test_file.py"
    with open(file_name, "w") as f:
        f.write("# Commit me!")

    command = Command.create_command("commit")
    command.apply("commit", "test_file committed")
    assert subprocess.check_output(["git", "diff", "--name-only"], text=True) == ""



import json
import logging
from importlib import resources
from json import JSONDecodeError
from pathlib import Path
from typing import Any, cast

from termcolor import cprint

mentat_dir_path = Path.home() / ".mentat"

# package_name should always be "mentat" - but this will work if package name is changed
package_name = __name__.split(".")[0]
default_config_file_name = "default_config.json"
config_file_name = ".mentat_config.json"
user_config_path = mentat_dir_path / config_file_name


class ConfigManager:
    def __init__(self, git_root: Path):
        self.git_root = git_root

        if user_config_path.exists():
            with open(user_config_path) as config_file:
                try:
                    self.user_config = json.load(config_file)
                except JSONDecodeError:
                    logging.info("User config file contains invalid json")
                    cprint(
                        "Warning: User .mentat_config.json contains invalid"
                        " json; ignoring user configuration file",
                        "light_yellow",
                    )
                    self.user_config = dict[str, str]()
        else:
            self.user_config = dict[str, str]()

        project_config_path = self.git_root / config_file_name
        if project_config_path.exists():
            with open(project_config_path) as config_file:
                try:
                    self.project_config = json.load(config_file)
                except JSONDecodeError:
                    logging.info("Project config file contains invalid json")
                    cprint(
                        "Warning: Git project .mentat_config.json contains invalid"
                        " json; ignoring project configuration file",
                        "light_yellow",
                    )
                    self.project_config = dict[str, str]()
        else:
            self.project_config = dict[str, str]()

        default_config_path = resources.files(package_name).joinpath(
            default_config_file_name
        )
        with default_config_path.open("r") as config_file:
            self.default_config = json.load(config_file)

    def input_style(self) -> list[tuple[str, str]]:
        return cast(list[tuple[str, str]], self._get_key("input-style"))

    def allow_32k(self) -> bool:
        return cast(bool, self._get_key("allow-32k"))

    def file_exclude_glob_list(self) -> list[str]:
        return cast(list[str], self._get_key("file-exclude-glob-list"))

    def _get_key(self, key: str) -> Any:
        if key in self.project_config:
            return self.project_config[key]
        elif key in self.user_config:
            return self.user_config[key]
        elif key in self.default_config:
            return self.default_config[key]
        else:
            raise ValueError(f"No value for config key {key} found")



import logging

from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.styles import Style
from termcolor import cprint

from .code_context import CodeContext
from .config_manager import ConfigManager
from .mentat_prompt_session import MentatPromptSession


class FilteredFileHistory(FileHistory):
    def __init__(self, filename: str, config: ConfigManager):
        self.excluded_phrases = ["y", "n", "i", "q"]
        super().__init__(filename)

    def append_string(self, string: str):
        if string.strip().lower() not in self.excluded_phrases:
            super().append_string(string)


class UserQuitInterrupt(Exception):
    pass


class UserInputManager:
    def __init__(self, config: ConfigManager, code_context: CodeContext):
        self.mentat_session = MentatPromptSession(
            code_context,
            message=[("class:prompt", ">>> ")],
            style=Style(config.input_style()),
        )
        # Won't have suggestions, completions, commands, etc.
        self.plain_session = PromptSession[str](
            message=[("class:prompt", ">>> ")],
            style=Style(config.input_style()),
        )

    def collect_user_input(self, use_plain_session: bool = False) -> str:
        if use_plain_session:
            user_input = self.plain_session.prompt()
        else:
            user_input = self.mentat_session.prompt()
        logging.debug(f"User input:\n{user_input}")
        user_input = user_input.strip()
        if user_input.lower() == "q":
            raise UserQuitInterrupt()
        return user_input

    def ask_yes_no(self, default_yes: bool) -> bool:
        cprint("(Y/n)" if default_yes else "(y/N)")
        while (
            user_input := self.collect_user_input(use_plain_session=True).lower()
        ) not in [
            "y",
            "n",
            "",
        ]:
            cprint("(Y/n)" if default_yes else "(y/N)")
        return user_input == "y" or (user_input != "n" and default_yes)



import logging
import os
import shlex
from collections import defaultdict
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, DefaultDict, Dict, Set, cast

from prompt_toolkit import PromptSession
from prompt_toolkit.application.current import get_app
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory, Suggestion
from prompt_toolkit.buffer import Buffer
from prompt_toolkit.completion import CompleteEvent, Completer, Completion
from prompt_toolkit.completion.word_completer import WordCompleter
from prompt_toolkit.document import Document
from prompt_toolkit.filters import Condition
from prompt_toolkit.formatted_text import AnyFormattedText
from prompt_toolkit.history import FileHistory
from prompt_toolkit.key_binding import KeyBindings
from prompt_toolkit.key_binding.key_processor import KeyPressEvent
from pygments.lexer import Lexer
from pygments.lexers import guess_lexer_for_filename
from pygments.token import Token
from pygments.util import ClassNotFound
from typing_extensions import override

from .code_context import CodeContext
from .commands import Command
from .config_manager import mentat_dir_path

logger = logging.getLogger()


class FilteredFileHistory(FileHistory):
    def __init__(self, filename: str):
        self.excluded_phrases = ["y", "n", "i", "q"]
        super().__init__(filename)

    def append_string(self, string: str):
        if (
            string.strip().lower() not in self.excluded_phrases
            # If the user mistypes a command, we don't want it to appear later
            and string.strip()
            and string.strip()[0] != "/"
        ):
            super().append_string(string)


class FilteredHistorySuggestions(AutoSuggestFromHistory):
    def __init__(self):
        super().__init__()

    def get_suggestion(self, buffer: Buffer, document: Document) -> Suggestion | None:
        # We want the auto completer to handle commands instead of the suggester
        if buffer.text[0] == "/":
            return None
        else:
            return super().get_suggestion(buffer, document)


@dataclass
class SyntaxCompletion:
    words: Set[str]
    created_at: datetime = datetime.utcnow()


class MentatCompleter(Completer):
    def __init__(self, code_context: CodeContext):
        self.code_context = code_context

        self.syntax_completions: Dict[Path, SyntaxCompletion] = dict()
        self.file_name_completions: DefaultDict[str, Set[Path]] = defaultdict(set)
        self.command_completer = WordCompleter(
            words=Command.get_command_completions(),
            ignore_case=True,
            sentence=True,
        )

        self._all_syntax_words: Set[str]
        self._last_refresh_at: datetime

        self.refresh_completions()

    def refresh_completions_for_file_path(self, file_path: Path):
        """Add/edit/delete completions for some filepath"""
        try:
            with open(file_path, "r") as f:
                file_content = f.read()
        except (FileNotFoundError, NotADirectoryError):
            logging.debug(f"Skipping {file_path}. Reason: file not found")
            return
        try:
            lexer = guess_lexer_for_filename(file_path, file_content)
            lexer = cast(Lexer, lexer)
        except ClassNotFound:
            logging.debug(f"Skipping {file_path}. Reason: lexer not found")
            return

        self.file_name_completions[file_path.name].add(file_path)

        tokens = list(lexer.get_tokens(file_content))
        filtered_tokens = set[str]()
        for token_type, token_value in tokens:
            if token_type not in Token.Name:
                continue
            if len(token_value) <= 1:
                continue
            filtered_tokens.add(token_value)
        self.syntax_completions[file_path] = SyntaxCompletion(words=filtered_tokens)

    def refresh_completions(self):
        # Remove syntax completions for files not in the context
        for file_path in set(self.syntax_completions.keys()):
            if file_path not in self.code_context.files:
                del self.syntax_completions[file_path]
                file_name = file_path.name
                self.file_name_completions[file_name].remove(file_path)
                if len(self.file_name_completions[file_name]) == 0:
                    del self.file_name_completions[file_name]

        # Add/update syntax completions for files in the context
        for file_path in self.code_context.files:
            if file_path not in self.syntax_completions:
                self.refresh_completions_for_file_path(file_path)
            else:
                modified_at = datetime.utcfromtimestamp(os.path.getmtime(file_path))
                if self.syntax_completions[file_path].created_at < modified_at:
                    self.refresh_completions_for_file_path(file_path)

        # Build de-duped syntax completions
        _all_syntax_words = set[str]()
        for syntax_completion in self.syntax_completions.values():
            _all_syntax_words.update(syntax_completion.words)
        self._all_syntax_words = _all_syntax_words

        self._last_refresh_at = datetime.utcnow()

    def get_completions(self, document: Document, complete_event: CompleteEvent):
        if document.text_before_cursor == "":
            return
        if (datetime.utcnow() - self._last_refresh_at).seconds > 5:
            self.refresh_completions()
        if (
            document.text_before_cursor[0] == "/"
            and not document.text_before_cursor[-1].isspace()
        ):
            command_completions = self.command_completer.get_completions(
                document, complete_event
            )
            for completion in command_completions:
                yield completion

        if document.text_before_cursor[-1] == " ":
            return
        document_words = document.text_before_cursor.split()
        if not document_words:
            return

        last_word = document_words[-1]
        get_completion_insert: Callable[[str], str] = lambda word: f"`{word}`"
        if last_word[0] == "`" and len(last_word) > 1:
            last_word = last_word.lstrip("`")
            get_completion_insert: Callable[[str], str] = lambda word: f"{word}`"

        completions = self._all_syntax_words.union(set(self.file_name_completions))

        for completion in completions:
            if completion.lower().startswith(last_word.lower()):
                file_names = self.file_name_completions.get(completion)
                if file_names:
                    for file_name in file_names:
                        yield Completion(
                            get_completion_insert(str(file_name)),
                            start_position=-len(last_word),
                            display=str(file_name),
                        )
                else:
                    yield Completion(
                        get_completion_insert(completion),
                        start_position=-len(last_word),
                        display=completion,
                    )


class MentatPromptSession(PromptSession[str]):
    def __init__(self, code_context: CodeContext, *args: Any, **kwargs: Any):
        self._setup_bindings()
        super().__init__(
            completer=MentatCompleter(code_context),
            history=FilteredFileHistory(str(mentat_dir_path / "history")),
            auto_suggest=FilteredHistorySuggestions(),
            multiline=True,
            prompt_continuation=self.prompt_continuation,
            key_bindings=self.bindings,
            *args,
            **kwargs,
        )

    @override
    def prompt(self, *args: Any, **kwargs: Any) -> str:
        # Automatically capture all commands
        while (user_input := super().prompt(*args, **kwargs)).startswith("/"):
            arguments = shlex.split(user_input[1:])
            command = Command.create_command(arguments[0])
            command.apply(*arguments[1:])
        return user_input

    @override
    def prompt_continuation(
        self, width: int, line_number: int, is_soft_wrap: int
    ) -> AnyFormattedText:
        return (
            "" if is_soft_wrap else [("class:continuation", " " * (width - 2) + "> ")]
        )

    def _setup_bindings(self):
        self.bindings = KeyBindings()

        @self.bindings.add("s-down")
        @self.bindings.add("c-j")
        def _(event: KeyPressEvent):
            event.current_buffer.insert_text("\n")

        @self.bindings.add("enter")
        def _(event: KeyPressEvent):
            event.current_buffer.validate_and_handle()

        @Condition
        def complete_suggestion() -> bool:
            app = get_app()
            return (
                app.current_buffer.suggestion is not None
                and len(app.current_buffer.suggestion.text) > 0
                and app.current_buffer.document.is_cursor_at_the_end
                and bool(app.current_buffer.text)
                and app.current_buffer.text[0] != "/"
            )

        @self.bindings.add("right", filter=complete_suggestion)
        def _(event: KeyPressEvent):
            suggestion = event.current_buffer.suggestion
            if suggestion:
                event.current_buffer.insert_text(suggestion.text)

        @self.bindings.add("c-c")
        def _(event: KeyPressEvent):
            if event.current_buffer.text != "":
                event.current_buffer.reset()
            else:
                event.app.exit(result="q")



import logging
import math
import os
from collections import defaultdict
from pathlib import Path
from typing import Union

from termcolor import cprint

from .change_conflict_resolution import (
    resolve_insertion_conflicts,
    resolve_non_insertion_conflicts,
)
from .code_change import CodeChange, CodeChangeAction
from .code_context import CodeContext
from .code_file import CodeFile
from .config_manager import ConfigManager
from .errors import MentatError
from .git_handler import get_git_diff_for_path
from .user_input_manager import UserInputManager


class CodeFileManager:
    def __init__(
        self,
        user_input_manager: UserInputManager,
        config: ConfigManager,
        code_context: CodeContext,
    ):
        self.user_input_manager = user_input_manager
        self.config = config
        self.code_context = code_context

    def _read_file(self, file: Union[Path, CodeFile]) -> list[str]:
        if isinstance(file, CodeFile):
            rel_path = file.path
        else:
            rel_path = self.config.git_root / file
        abs_path = self.config.git_root / rel_path

        with open(abs_path, "r") as f:
            lines = f.read().split("\n")
        return lines

    def _read_all_file_lines(self) -> None:
        self.file_lines = dict[Path, list[str]]()
        for file in self.code_context.files.values():
            rel_path = Path(os.path.relpath(file.path, self.config.git_root))
            # here keys are str not path object
            self.file_lines[rel_path] = self._read_file(file)

    def get_code_message(self):
        self._read_all_file_lines()
        code_message = ["Code Files:\n"]
        for file in self.code_context.files.values():
            abs_path = file.path
            rel_path = Path(os.path.relpath(abs_path, self.config.git_root))

            # We always want to give GPT posix paths
            posix_rel_path = Path(rel_path).as_posix()
            code_message.append(posix_rel_path)

            for i, line in enumerate(self.file_lines[rel_path], start=1):
                if file.contains_line(i):
                    code_message.append(f"{i}:{line}")
            code_message.append("")

            git_diff_output = get_git_diff_for_path(self.config.git_root, rel_path)
            if git_diff_output:
                code_message.append("Current git diff for this file:")
                code_message.append(f"{git_diff_output}")

        return "\n".join(code_message)

    def _add_file(self, abs_path: Path):
        logging.info(f"Adding new file {abs_path} to context")
        self.code_context.files[abs_path] = CodeFile(abs_path)
        # create any missing directories in the path
        abs_path.parent.mkdir(parents=True, exist_ok=True)

    def _delete_file(self, abs_path: Path):
        logging.info(f"Deleting file {abs_path}")
        if abs_path in self.code_context.files:
            del self.code_context.files[abs_path]
        abs_path.unlink()

    def _handle_delete(self, delete_change: CodeChange):
        abs_path = self.config.git_root / delete_change.file
        if not abs_path.exists():
            logging.error(f"Path {abs_path} non-existent on delete")
            return

        cprint(f"Are you sure you want to delete {delete_change.file}?", "red")
        if self.user_input_manager.ask_yes_no(default_yes=False):
            cprint(f"Deleting {delete_change.file}...")
            self._delete_file(abs_path)
        else:
            cprint(f"Not deleting {delete_change.file}")

    def _get_new_code_lines(
        self, rel_path: Path, changes: list[CodeChange]
    ) -> list[str] | None:
        if not changes:
            return []
        if len(set(map(lambda change: change.file, changes))) > 1:
            raise Exception("All changes passed in must be for the same file")

        changes = sorted(changes, reverse=True)

        # We resolve insertion conflicts twice because non-insertion conflicts
        # might move insert blocks outside of replace/delete blocks and cause
        # them to conflict again
        changes = resolve_insertion_conflicts(changes, self.user_input_manager, self)
        changes = resolve_non_insertion_conflicts(changes, self.user_input_manager)
        changes = resolve_insertion_conflicts(changes, self.user_input_manager, self)
        if not changes:
            return []

        new_code_lines = self.file_lines[rel_path].copy()
        if new_code_lines != self._read_file(rel_path):
            logging.info(f"File '{rel_path}' changed while generating changes")
            cprint(
                f"File '{rel_path}' changed while generating; current file changes"
                " will be erased. Continue?",
                color="light_yellow",
            )
            if not self.user_input_manager.ask_yes_no(default_yes=False):
                cprint(f"Not applying changes to file {rel_path}.")
                return None

        # Necessary in case the model needs to insert past the end of the file
        last_line = len(new_code_lines) + 1
        largest_changed_line = math.ceil(changes[0].last_changed_line)
        if largest_changed_line > last_line:
            new_code_lines += [""] * (largest_changed_line - last_line)

        min_changed_line = largest_changed_line + 1
        for change in changes:
            if change.last_changed_line >= min_changed_line:
                raise MentatError(f"Change line number overlap in file {change.file}")
            min_changed_line = change.first_changed_line
            new_code_lines = change.apply(new_code_lines)
        return new_code_lines

    def write_changes_to_files(self, code_changes: list[CodeChange]) -> None:
        file_changes = defaultdict[Path, list[CodeChange]](list)
        for code_change in code_changes:
            rel_path = code_change.file
            abs_path = self.config.git_root / rel_path
            match code_change.action:
                case CodeChangeAction.CreateFile:
                    cprint(f"Creating new file {rel_path}", color="light_green")
                    self._add_file(abs_path)
                    with open(abs_path, "w") as f:
                        f.write("\n".join(code_change.code_lines))
                case CodeChangeAction.DeleteFile:
                    self._handle_delete(code_change)
                case CodeChangeAction.RenameFile:
                    abs_new_path = self.config.git_root / code_change.name
                    self._add_file(abs_new_path)
                    code_lines = self.file_lines[rel_path]
                    with open(abs_new_path, "w") as f:
                        f.write("\n".join(code_lines))
                    self._delete_file(abs_path)
                    file_changes[code_change.name] += file_changes[rel_path]
                    file_changes[rel_path] = []
                    self.file_lines[code_change.name] = self._read_file(abs_new_path)
                case _:
                    file_changes[rel_path].append(code_change)

        for rel_path, changes in file_changes.items():
            abs_path = self.config.git_root / rel_path
            new_code_lines = self._get_new_code_lines(rel_path, changes)
            if new_code_lines:
                if abs_path not in self.code_context.files:
                    raise MentatError(
                        f"Attempted to edit file {abs_path} not in context"
                    )
                with open(abs_path, "w") as f:
                    f.write("\n".join(new_code_lines))



import os
from textwrap import dedent

from mentat.code_context import CodeContext
from mentat.code_file_manager import CodeFileManager
from mentat.user_input_manager import UserInputManager


# Make sure we always give posix paths to GPT
def test_posix_paths(mock_config):
    dir_name = "dir"
    file_name = "file.txt"
    file_path = os.path.join(dir_name, file_name)
    os.makedirs(dir_name, exist_ok=True)
    with open(file_path, "w") as file_file:
        file_file.write("I am a file")
    code_context = CodeContext(
        config=mock_config,
        paths=[file_path],
        exclude_paths=[],
    )
    code_file_manager = CodeFileManager(
        user_input_manager=UserInputManager(
            config=mock_config, code_context=code_context
        ),
        config=mock_config,
        code_context=code_context,
    )
    code_message = code_file_manager.get_code_message()
    assert dir_name + "/" + file_name in code_message.split("\n")


def test_partial_files(mock_config):
    dir_name = "dir"
    file_name = "file.txt"
    file_path = os.path.join(dir_name, file_name)
    os.makedirs(dir_name, exist_ok=True)
    with open(file_path, "w") as file_file:
        file_file.write(dedent("""\
             I am a file
             with 5 lines
             third
             fourth
             fifth"""))
    file_path_partial = file_path + ":1,3-5"

    code_context = CodeContext(
        config=mock_config,
        paths=[file_path_partial],
        exclude_paths=[],
    )
    code_file_manager = CodeFileManager(
        user_input_manager=UserInputManager(
            config=mock_config, code_context=code_context
        ),
        config=mock_config,
        code_context=code_context,
    )
    code_message = code_file_manager.get_code_message()
    assert code_message == dedent("""\
             Code Files:

             dir/file.txt
             1:I am a file
             3:third
             4:fourth
             5:fifth
              """)



import os
from textwrap import dedent

from mentat.app import run


def test_system(mock_call_llm_api, mock_collect_user_input, mock_setup_api_key):
    # Create a temporary file
    temp_file_name = "temp.py"
    with open(temp_file_name, "w") as f:
        f.write("# This is a temporary file.")

    mock_collect_user_input.side_effect = [
        'Replace comment with print("Hello, world!")',
        "y",
        KeyboardInterrupt,
    ]
    mock_call_llm_api.set_generator_values([dedent("""\
        I will add a print statement.

        Steps:
        1. Add a print statement after the last line

        @@start
        {{
            "file": "{file_name}",
            "action": "replace",
            "start-line": 1,
            "end-line": 1
        }}
        @@code
        print("Hello, world!")
        @@end""".format(file_name=temp_file_name))])

    # Run the system with the temporary file path
    run([temp_file_name])

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = 'print("Hello, world!")'
    assert content == expected_content


def test_interactive_change_selection(
    mock_call_llm_api, mock_collect_user_input, mock_setup_api_key
):
    # Create a temporary file
    temp_file_name = "temp_interactive.py"
    with open(temp_file_name, "w") as f:
        f.write("# This is a temporary file for interactive test.")

    mock_collect_user_input.side_effect = [
        "Add changes to the file",
        "i",
        "y",
        "n",
        "y",
        KeyboardInterrupt,
    ]
    mock_call_llm_api.set_generator_values([dedent("""\
        I will make three changes to the file.

        Steps:
        1. Replace the comment with print("Change 1")
        2. Add print("Change 2") after the first line
        3. Add print("Change 3") after the second line

        @@start
        {{
            "file": "{file_name}",
            "action": "replace",
            "start-line": 1,
            "end-line": 1
        }}
        @@code
        print("Change 1")
        @@end
        @@start
        {{
            "file": "{file_name}",
            "action": "insert",
            "insert-after-line": 1,
            "insert-before-line": 2
        }}
        @@code
        print("Change 2")
        @@end
        @@start
        {{
            "file": "{file_name}",
            "action": "insert",
            "insert-after-line": 2,
            "insert-before-line": 3
        }}
        @@code
        print("Change 3")
        @@end""".format(file_name=temp_file_name))])

    # Run the system with the temporary file path
    run([temp_file_name])

    # Check if the temporary file is modified as expected
    with open(temp_file_name, "r") as f:
        content = f.read()
        expected_content = 'print("Change 1")\n\nprint("Change 3")'
    assert content == expected_content


# Makes sure we're properly turning the model output into correct path no matter the os
def test_without_os_join(
    mock_call_llm_api, mock_collect_user_input, mock_setup_api_key
):
    temp_dir = "dir"
    temp_file_name = "temp.py"
    temp_file_path = os.path.join(temp_dir, temp_file_name)
    os.makedirs(temp_dir, exist_ok=True)
    with open(temp_file_path, "w") as f:
        f.write("# This is a temporary file.")

    mock_collect_user_input.side_effect = [
        'Replace comment with print("Hello, world!")',
        "y",
        KeyboardInterrupt,
    ]

    # Use / here since that should always be what the model outputs
    fake_file_path = temp_dir + "/" + temp_file_name
    mock_call_llm_api.set_generator_values([dedent("""\
        I will add a print statement.

        Steps:
        1. Add a print statement after the last line

        @@start
        {{
            "file": "{file_name}",
            "action": "replace",
            "start-line": 1,
            "end-line": 1
        }}
        @@code
        print("Hello, world!")
        @@end""".format(file_name=fake_file_path))])

    run([temp_file_path])
    with open(temp_file_path, "r") as f:
        content = f.read()
        expected_content = 'print("Hello, world!")'
    assert content == expected_content
